### Python dataset class, agnostic to backend


import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
import typing
from typing import Sequence, Union, Tuple, Optional, Generator, Callable
import numpy
import datetime
import warnings


@dataclass
class EventInfo:
    """ Pure python event information. In effect duplicating the most important bits of the ROOT header"""
    eventNumber: int
    station : int
    run: int
    readoutTime : float
    triggerTime : float
    triggerType: str
    sysclk: int
    sysclkLastPPS: Tuple[int, int]  # the last 2 PPS sysclks, most recent first
    pps: int
    radiantStartWindows: numpy.ndarray
    sampleRate: Optional[float]  # Sample rate, in GSa/s
    radiantThrs: Optional[numpy.ndarray]
    lowTrigThrs: Optional[numpy.ndarray]

##
class AbstractDataset(ABC):
    """ Abstract Base Class for accessing RNO-G data in Python , implemented either with uproot or PyROOT"""

    def setEntries(self, i : Union[int, Tuple[int, int]]):
        """
        Select entries to read out with wfs or eventInfo. Can either be a
        single entry or a tuple representating start and end entries, exclusive.
        Use (0,dataset.N()) or (0, None) to select all events. Negative indices indicate from end.
        """

        if isinstance(i, tuple):
            self.multiple = True
            self.first = i[0]
            self.last = i[1]
            if self.last is None:
                self.last = self.N()
            if self.first < 0:
                self.first += self.N()
            if self.last < 0:
                self.last += self.N()

            if self.last > self.N():
                self.last = self.N()
        else:
            self.multiple = False
            if i < 0:
                i+= self.N() 
            if i is None: 
                i = 0 
            self.entry = i
            self.first = i
            self.last = i + 1


    def N(self) -> int:
        """ Return the number of events available in this dataset"""
        return 0


    @abstractmethod
    def _iterate(self, start: int , stop : int , calibrated: bool, max_entries_in_mem: int, selector: Optional[Callable[[EventInfo],bool]]) -> Generator[Tuple[Optional[EventInfo], Optional[numpy.ndarray]],None,None]:
        """ implementation-defined part of iterator"""
        pass

    def iterate(self, start : int = 0, stop : Union[int,None] = None,
                calibrated: bool = False, max_entries_in_mem : int = 256,
                selector: Optional[Callable[[EventInfo], bool]] = None) \
                -> Generator[Tuple[Optional[EventInfo], Optional[numpy.ndarray]], None, None]:
        """ Iterate over events from start to stop, holding at most max_entries_in_mem in RAM.
            Returns a tuple of EventInfo and the event waveforms (potentially calibrated).
        """
        if start < 0:
            start += self.N()
        if start < 0 or start > self.N():
            return

        if stop is None:
            stop = self.N()

        if stop < 0:
            stop += self.N()

        if stop < 0 or start > self.N():
            return

        yield from self._iterate(start, stop, calibrated, max_entries_in_mem, selector)

    @abstractmethod
    def eventInfo(self) -> Union[Optional[EventInfo], Sequence[Optional[EventInfo]]]:
        """Get selected event info(s)
           Depending on what was passed to setEntries this can return either one EventInfo or a list of them
        """
        pass

    @abstractmethod
    def wfs(self, calibrated : bool = False) -> Optional[numpy.ndarray]:
        """ Get select waveform(s).
            Depending on what was passed to setEntries, this may be a single waveform or many
        """
        pass
    
    def get_selected_wfs(self, selector: Callable[[EventInfo], bool], calibrated : bool = False) -> Optional[numpy.ndarray]:
        """ Convenience interface to use selector """
        return numpy.array([wf for _, wf in self.iterate(start=self.first, stop=self.last, selector=selector)])


def Dataset(station : int = 0, run : int = 0, data_path : Optional[str] = None, backend : str= "auto", 
            verbose : bool = False, skip_incomplete : bool = True,
            read_daq_status : bool = True, read_run_info : bool = True,
            preferred_file : Optional[str] = None, 
            *, data_dir : Optional[str] = None ) -> Optional[AbstractDataset]:
   """

   This is not a class, but a factory method! This is meant to be the interface
   to rule them all for loading RNO-G data. Due to Cosmin's poor initial API
   design, it has become perhaps more complicated than it should be. 

   In the case of setting station = 0 and run = 0, data_path will be
   interpreted as a directory containing ROOT files, which is useful if you
   don't have the full directory hierarchy setup or want to look at data taken
   with the fakedaq.

   If data_path is a file rather than directory, then that file will be
   attempted to be loaded as a combined file, ignoring the provided run and
   station numbers. 

   If data_path is a directory and station/run are non-zero, this returns a
   dataset corresponding to the station and run using data_path as the base
   directory (i.e. the folder hierarcy is structured something like
   ${data_path}/stationX/runY/*.root). 


   If data_path is None (or ""), then the environmental variable RNO_G_DATA
   (and also  RNO_G_ROOT_DATA) will be queried.  data_path can also be a URL
   for loading of files via HTTP (e.g.
   https://user:password@example.com/rno-g-data), though there may be some
   subtleties about escaping passwords that may differ betweeen different
   backends. 

   The backend can be chosen explicitly ("pyroot" or "uproot") or auto will try
   to use the best one ("pyroot" if available, otherwise reverting to "uproot").

   verbose prints out things mostly useful for debugging.

   read_daq_status and read_run_info are mostly self-explanatory. Avoiding
   reading them may speed things up or work around the files not being there
   for some reason. 

   skip_incomplete affects what happens when a dataset is incomplete
   (telemetered). If True, will only index fully telemetered events. If False,
   will be indexed by full dataset, but untelemetered events will have
   waveforms of None type (if requested singly) or be all 0's (if requested
   via the bulk interface, as numpy doesn't support jagged ararys).

   preferred_file, if not None or "", and data_path is not a file, will further
   change the loading behavior. By default,  we will try to load full waveforms
   falling back to loading combined.root. But if preferred_file is set, it will
   prefer loading ${preferred_file}.root if possible, treating it as a file in
   the same format as combined.root.  For example, you can set it to "combined"
   to load combined.root even if full waveforms are available. Or, if you have
   your own subselection in the same format (e.g. for example if you generated
   a file that is only forced triggers) this provides an arguably convenient
   way to load those. 

   """


   # handle deprecated name data_dir 
   if data_dir is not None: 
       warnings.warn("data_dir is deprecated, use data_path instead. This may be removed in the future, breaking your code.")
       if data_path is not None: 
           raise TypeError("Dataset received both data_path and data_dir!") 
       data_path = data_dir 


   #treat "" as an alias for None
   if data_path == "": 
       data_path = None


   if data_path is None:
       for env_var in ['RNO_G_DATA', 'RNO_G_ROOT_DATA']: 
           if env_var in os.environ: 
               data_path = os.environ[env_var] 
               break 
       if data_path is None:                
           print("Neither data_path nor any relevant environmental variable (e.g. RNO_G_DATA) is defined and I don't know where else to look :(")
           return None 

   if backend == "auto":
        try:
            import ROOT
            import mattak.backends.pyroot.mattakloader
            if verbose:
                print('Using pyroot backend')
            backend = "pyroot"
        except:
            try:
                import uproot
                backend = "uproot"
                if verbose:
                    print('Using uproot backend')
            except:
                print("No backends available")
                return None

   if backend == "uproot":
        import mattak.backends.uproot.dataset
        return mattak.backends.uproot.dataset.Dataset(station, run, data_path, verbose, skip_incomplete, read_daq_status, read_run_info, preferred_file)
   elif backend == "pyroot":
        import mattak.backends.pyroot.dataset
        return mattak.backends.pyroot.dataset.Dataset(station, run, data_path, verbose, skip_incomplete, read_daq_status, read_run_info, preferred_file)
   else:
       print("Unknown backend (known backends are \"uproot\" and \"pyroot\")")
       return None



import os 
import uproot
import errno 
import os.path
import configparser
import mattak.Dataset
import typing
from typing import Union, List, Optional, Tuple, Generator, Callable, Sequence
import numpy
import math

# Dublicated from Dataset.cc
waveform_tree_names = ["waveforms", "wfs", "wf", "waveform"]
header_tree_names = ["hdr", "header", "hd", "hds", "headers"]
daqstatus_tree_names = ["daqstatus", "ds", "status"]

def read_tree(ur_file, tree_names):
    """
    Parameters
    ----------
    
    ur_file: uproot.open()
        Input file
        
    tree_names: list(str)
        List of possible tree names, the first found is returned
        
    Returns
    -------
    
    tree: uproot.tree    
    """
    
    for tree_name in tree_names:
        if tree_name in ur_file:
            tree = ur_file[tree_name]
            return tree, tree_name


class Dataset(mattak.Dataset.AbstractDataset):

    def __init__(self, station : int, run : int, data_path : str, verbose : bool = False,
                 skip_incomplete : bool = True, read_daq_status : bool = True,
                 read_run_info : bool = True, file_preference: Optional[str] = None):
        
        self.backend = "uproot"
        self.__verbose = verbose
        self.__read_daq_status = read_daq_status
        self.__read_run_info = read_run_info

        #special case where data_path is a file

        if os.path.isfile(data_path): 
            self.data_path_is_file = True
            self.rundir = data_path 
        else: 
            self.data_path_is_file = False
            # special case where we load a directory instead of a station/run
            if station == 0 and run == 0: 
                self.rundir = data_path
            else: 
                self.rundir = "%s/station%d/run%d" % (data_path,station,run)



        if (skip_incomplete == False  and self.data_path_is_file):
            print("skip_incomplete = false is incompatible with data_path as file"); 
            skip_incomplete = True 

        self.skip_incomplete = skip_incomplete 

        # this duplicates a bunch of C++ code in mattak::Dataset
        # check for full or partial run by looking for waveforms.root
        # Only open files/trees, do not access data

        self.combined_tree = None 

        # check for a preference
        if (file_preference is not None and file_preference != "") or self.data_path_is_file: 
            preferred_file = "%s:combined" % (self.rundir) if self.data_path_is_file  else "%s/%s.root:combined" %(self.rundir, file_preference) 
            try: 
                self.combined_tree = uproot.open(preferred_file)
                if verbose: 
                    print ("Found preferred file %s" % (preferred_file))
                    print (self.combined_tree)
                self.full = False
            except Exception:
                # can't recover from this 
                if self.data_path_is_file: 
                    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self.rundir)
                print ("Could not find preferred file %s. Falling back to normal behavior" % (preferred_file))


        

        #if we didn't load the combined_tree already , try to load full tree 
        if self.combined_tree is None:
            try:
                self.wf_file = uproot.open("%s/waveforms.root" % (self.rundir))
                if self.__verbose:
                    print ("Open waveforms.root (Found full run folder) ...") 

                self.combined_tree = None
                self.full = True

                self.wf_tree, self.wf_branch = read_tree(self.wf_file, waveform_tree_names)
                self._wfs = self.wf_tree[self.wf_branch]

                self.hd_file = uproot.open("%s/headers.root" % (self.rundir))
                self.hd_tree,self.hd_branch = read_tree(self.hd_file, header_tree_names); 
                self._hds = self.hd_tree[self.hd_branch]

                if self.__read_daq_status:
                    self.ds_file = uproot.open("%s/daqstatus.root" % (self.rundir))
                    self.ds_tree, self.ds_branch = read_tree(self.ds_file, daqstatus_tree_names)
                    self._dss = self.ds_tree[self.ds_branch]
            except Exception: 
                self.full = False

        # we haven't already loaded the full tree
        if not self.full:   
            if self.combined_tree is None: # we didn't already load our preference
                self.combined_tree = uproot.open("%s/combined.root:combined" %(self.rundir))
                if self.__verbose: 
                    print ("Found combined file")
                    print (self.combined_tree)

            self._wfs, self.wf_branch = read_tree(self.combined_tree, waveform_tree_names)

            # build an index of the waveforms we do have
            if not self.skip_incomplete: 
                wfs_included = self._wfs['event_number'].array() 
                self.events_with_waveforms = {ev: idx for idx, ev in enumerate(wfs_included)}

                self.full_head_file = uproot.open("%s/headers.root" % (self.rundir))
                self.full_head_tree,_ = read_tree(self.full_head_file, header_tree_names)
                
                if self.__read_daq_status:
                    self.full_daq_file = uproot.open("%s/daqstatus.root" % (self.rundir))
                    self.full_daq_tree, _ = read_tree(self.full_daq_file, daqstatus_tree_names)


            # Get header and daq information from combined file or from full run
            hd_tree = self.combined_tree if skip_incomplete else self.full_head_tree
            self._hds, self.hd_branch = read_tree(hd_tree, header_tree_names)
                
            if self.__read_daq_status:
                ds_tree = self.combined_tree if skip_incomplete else self.full_daq_tree
                self._dss, self.ds_branch =  read_tree(ds_tree, daqstatus_tree_names) 

        if station == 0 and run == 0 or self.data_path_is_file: 
            self.station = self._hds['station_number'].array(entry_start=0, entry_stop=1)[0]
            self.run = self._hds['run_number'].array(entry_start=0, entry_stop=1)[0]

        else:
            self.station = station
            self.run = run

        self.data_path = data_path
        self.setEntries(0) 

        self.run_info = None
        if self.__read_run_info:
            # try to get the run info, if we're using combined tree, try looking in there 
            # doh, uproot can't read the runinfo ROOT files... let's parse the text files instead
            
            if os.path.exists("%s/aux/runinfo.txt" % (self.rundir)): 
                with open("%s/aux/runinfo.txt" % (self.rundir)) as fruninfo: 
                    # we'll abuse configparser to read the runinfo, but we have to add a dummy 
                    # section to properly abuse it 
                    config = configparser.ConfigParser() 
                    config.read_string('[dummy]\n' + fruninfo.read())
                    self.run_info = config['dummy']
                

    def eventInfo(self) -> Union[Optional[mattak.Dataset.EventInfo],Sequence[Optional[mattak.Dataset.EventInfo]]]: 
        kw = dict(entry_start = self.first, entry_stop = self.last)
        
        station = self._hds['station_number'].array(**kw)
        run = self._hds['run_number'].array(**kw)
        eventNumber = self._hds['event_number'].array(**kw)
        readoutTime = self._hds['readout_time'].array(**kw)
        triggerTime = self._hds['trigger_time'].array(**kw)
        triggerInfo = self._hds['trigger_info'].array(**kw)
        pps = self._hds['pps_num'].array(**kw)
        sysclk = self._hds['sysclk'].array(**kw)
        sysclk_lastpps = self._hds['sysclk_last_pps'].array(**kw)
        sysclk_lastlastpps = self._hds['sysclk_last_last_pps'].array(**kw)
        
        if self.__read_daq_status:
            radiantThrs = numpy.array(self._dss['radiant_thresholds[24]'])
            lowTrigThrs = numpy.array(self._dss['lt_trigger_thresholds[4]'])
        
        if self.run_info is not None:
            sampleRate = float(self.run_info['radiant-samplerate']) / 1000
        else:
            sampleRate = None

        # um... yeah, that's obvious 
        radiantStartWindows = self._hds['trigger_info/trigger_info.radiant_info.start_windows[24][2]'].array(**kw, library='np')
        
        infos = [] 
        info = None  # if range(0)
        for i in range(self.last-self.first): 
            
            triggerType  = "UNKNOWN"; 
            if triggerInfo[i]['trigger_info.radiant_trigger']:
                which = triggerInfo[i]['trigger_info.which_radiant_trigger']
                if which == -1: 
                    which = "X" 
                triggerType = "RADIANT" + str(which)
            elif triggerInfo[i]['trigger_info.lt_trigger']:
                triggerType = "LT"
            elif triggerInfo[i]['trigger_info.force_trigger']:
                triggerType = "FORCE"
            elif triggerInfo[i]['trigger_info.pps_trigger']:
                triggerType = "PPS"

            info = mattak.Dataset.EventInfo(eventNumber = eventNumber[i], 
                                            station = station[i],
                                            run = run[i],
                                            readoutTime = readoutTime[i],
                                            triggerTime = triggerTime[i],
                                            triggerType = triggerType,
                                            sysclk = sysclk[i],
                                            sysclkLastPPS = (sysclk_lastpps[i], sysclk_lastlastpps[i]), 
                                            pps = pps[i], 
                                            radiantStartWindows = radiantStartWindows[i],
                                            sampleRate = sampleRate,
                                            radiantThrs=radiantThrs[i] if self.__read_daq_status else None,
                                            lowTrigThrs=lowTrigThrs[i] if self.__read_daq_status else None)
            
            infos.append(info)
            
        if not self.multiple:
            return info
        else:
            return infos
    
    
    def N(self) -> int: 
        return self._hds.num_entries

    def wfs(self, calibrated : bool =False) -> Optional[numpy.ndarray]: 
        # assert(not calibrated) # not implemented yet 

        w = None 
        if self.full or self.skip_incomplete: 
            w = self._wfs['radiant_data[24][2048]'].array(entry_start=self.first, entry_stop=self.last, library='np')
        elif not self.multiple: 
            if self.first in self.events_with_waveforms: 
                idx = self.events_with_waveforms[self.first]
                w = self._wfs['radiant_data[24][2048]'].array(entry_start=idx, entry_stop=idx+1, library='np')
        else: 
            # so ... we need to loop through and find which things we have actually have waveforms
            # start by allocating the output
            w = numpy.zeros((self.last - self.first, 24, 2048), dtype='float64' if calibrated else 'int16')

            # now figure out how much of the data array we need 
            wf_start = None
            wf_end = None

            # store the indices that will be non-zero
            wf_idxs = [] 
            for i in range(self.first, self.last): 
                if i in self.events_with_waveforms: 
                    wf_idxs.append(i - self.first) 
                    # these are the start and stop of our array we need to load
                    if wf_start is None: 
                        wf_start = self.events_with_waveforms[i]
                    wf_end = self.events_with_waveforms[i] + 1 
            
            if len(wf_idxs): 
                w[wf_idxs] = self._wfs['radiant_data[24][2048]'].array(entry_start=wf_start, entry_stop=wf_end, library='np')

        # here we'd eventually handle calibration I think? 

        if self.multiple: 
            return w
        return None if w is None else w[0] 

    def _iterate(self, start : int , stop : int , calibrated: bool,  max_in_mem : int, 
                 selector: Optional[Callable[[mattak.Dataset.EventInfo],bool]] = None) -> Generator[Tuple[Optional[mattak.Dataset.EventInfo], Optional[numpy.ndarray]],None,None]:

       # cache current values given by setEntries(..)
        original_entry : Union[int, Tuple[int,int]] = (self.first, self.last) if self.multiple else self.entry

        # determine in how many batches we want to access the data given how much events we want to load into the RAM at once
        n_batches = math.ceil((stop - start) / max_in_mem)

        for i_batch in range(n_batches):

            # looping over the batches defining the start and stop index
            batch_start = start + i_batch * max_in_mem
            batch_stop = min(stop, batch_start + max_in_mem)
            
            self.setEntries((batch_start, batch_stop))
            
            # load events from file 
            w = self.wfs(calibrated)
            e = self.eventInfo()
            
            # we modified the internal data pointers with the prev. call of self.setEntries(...)
            # this is intransparent for the outside world and has to be reverted
            self.setEntries(original_entry)

            for idx in range(batch_stop - batch_start):
                if selector is not None:
                    if selector(e[idx]):
                        yield e[idx], w[idx]
                else:
                    yield e[idx], w[idx]




import ROOT
import mattak.backends.pyroot.mattakloader
import mattak.Dataset
from typing import Sequence, Union, Tuple, Optional, Callable, Generator
import numpy
import os.path 
import warnings

try:
    import cppyy.ll

# work around a weird issue that happens on some systems
# for some reason, in some configurations it  doesn't detect free in the global namespace.
# So we'll define a different function in the global namespace with a super creative name
# that does the same thing as free then set it equal to free.
except AttributeError:
    import cppyy
    cppyy.cppdef("void freee(void *p) { free(p); }")
    cppyy.gbl.free = cppyy.gbl.freee
    import cppyy.ll


cppyy.cppdef(" bool is_nully(void *p) { return !p; }");


def isNully(p):
    return p is None or ROOT.AddressOf(p) == 0 or  cppyy.gbl.is_nully(p)

class Dataset(mattak.Dataset.AbstractDataset):

    def __init__(self, station : int, run : int, data_path : str, 
                 verbose : bool = False, skip_incomplete : bool = True, 
                 read_daq_status : bool = True, read_run_info : bool = True, 
                 preferred_file : Optional[str] = None):
        
        self.backend = "pyroot"
        self.__read_daq_status = read_daq_status
        self.__read_run_info = read_run_info

        opt = ROOT.mattak.DatasetOptions()

        opt.partial_skip_incomplete = skip_incomplete
        opt.verbose = verbose 
        opt.base_data_dir = data_path # only has an effect if it's actually a dir

        if preferred_file is not None and preferred_file != "":
            opt.file_preference = preferred_file
        self.ds = ROOT.mattak.Dataset(opt)

        if data_path is not None and os.path.isfile(data_path):
            self.ds.loadCombinedFile(data_path)
        elif (station == 0 and run == 0):
            self.ds.loadDir(data_path)
        else:
            self.ds.loadRun(station,run)
                    
        if self.N() < 0: 
            raise IOError("Could not load run [data_path: %s, %d %d]" % (data_path, station, run)) 

        if self.N() == 0: 
            warnings.warn("Run is empty?")
            self.station = station
            self.run = run
        else: 
            self.station = self.ds.header().station_number
            self.run = self.ds.header().run_number

        self.data_path = data_path
        self.setEntries(0)

        if verbose: 
            print("We think we found station %d run %d" % (self.station,self.run))

    def N(self) -> int:
        return self.ds.N()

    def _eventInfo(self, i : int) -> Optional[mattak.Dataset.EventInfo]:
        #TODO: handle this in C++ code if it's too slow in Python
        if not self.ds.setEntry(i):
            return None
        
        hdr = self.ds.header()
        
        if self.__read_daq_status:
            daq_status = self.ds.status()
            radiantThrs = numpy.array(daq_status.radiant_thresholds)
            lowTrigThrs = numpy.array(daq_status.lt_trigger_thresholds)
        else:
            radiantThrs = None
            lowTrigThrs = None
            
        if self.__read_run_info:
            runinfo = self.ds.info()
            sampleRate = runinfo.radiant_sample_rate / 1000 if not isNully(runinfo) else 3.2 
        else:
            sampleRate = None

        assert(hdr.station_number == self.station)
        assert(hdr.run_number == self.run)
        
        station = hdr.station_number
        run = hdr.run_number
        eventNumber = hdr.event_number
        readoutTime = hdr.readout_time
        triggerTime = hdr.trigger_time
        triggerType = "UNKNOWN"
        
        if hdr.trigger_info.radiant_trigger:
            which = hdr.trigger_info.which_radiant_trigger
            if which == -1:
                which = "X"
            triggerType = "RADIANT" + str(which)
        elif hdr.trigger_info.lt_trigger:
            triggerType = "LT"
        elif hdr.trigger_info.force_trigger:
            triggerType = "FORCE"
        elif hdr.trigger_info.pps_trigger:
            triggerType = "PPS"

        pps = hdr.pps_num
        sysclk = hdr.sysclk
        sysclkLastPPS = (hdr.sysclk_last_pps, hdr.sysclk_last_last_pps)
        radiantStartWindows = numpy.frombuffer(cppyy.ll.cast['uint8_t*'](hdr.trigger_info.radiant_info.start_windows), dtype='uint8', count=24 * 2).reshape(24, 2)

        return mattak.Dataset.EventInfo(eventNumber = eventNumber,
                                        station = station,
                                        run = run,
                                        readoutTime=readoutTime,
                                        triggerTime=triggerTime,
                                        triggerType=triggerType,
                                        sysclk=sysclk,
                                        sysclkLastPPS=sysclkLastPPS,
                                        pps=pps,
                                        radiantStartWindows = radiantStartWindows,
                                        sampleRate = sampleRate,
                                        radiantThrs=radiantThrs,
                                        lowTrigThrs=lowTrigThrs)


    def eventInfo(self) -> Union[Optional[mattak.Dataset.EventInfo],Sequence[Optional[mattak.Dataset.EventInfo]]]:

        if self.multiple:
            infos = []
            for i in range(self.first, self.last):
                infos.append(self._eventInfo(i))
            
            return infos

        return self._eventInfo(self.entry)

    def _wfs(self, i : int, calibrated: bool = False):
        self.ds.setEntry(i)
        wf = self.ds.calibrated() if calibrated else self.ds.raw()
        if isNully(wf):
            return None
        
        return numpy.frombuffer(cppyy.ll.cast['double*' if calibrated else 'int16_t*'](wf.radiant_data), dtype = 'float64' if calibrated else 'int16', count=24 * 2048).reshape(24,2048)


    def wfs(self, calibrated : bool=False) -> Optional[numpy.ndarray]: 

        # the simple case first
        if not self.multiple:
            return self._wfs(self.entry, calibrated)

        if self.last - self.first < 0:
            return None

        out = numpy.zeros((self.last - self.first, 24, 2048), dtype='float64' if calibrated else 'int16')
        for entry in range(self.first, self.last):
            this_wfs = self._wfs(entry, calibrated)
            if this_wfs is not None:
                out[entry-self.first][:][:] = this_wfs

        return out

       
    def _iterate(self, start : int, stop : int, calibrated : bool , max_in_mem : int,
                 selector: Optional[Callable[[mattak.Dataset.EventInfo], bool]] = None) -> Generator[Tuple[Optional[mattak.Dataset.EventInfo], Optional[numpy.ndarray]],None,None]:

        if selector is not None:
            for i in range(start, stop):
                evinfo = self._eventInfo(i)
                if evinfo is not None and selector(evinfo):
                    yield evinfo, self._wfs(i, calibrated)
        else:
            for i in range(start, stop):
                yield self._eventInfo(i), self._wfs(i, calibrated)
        return







import ROOT 
import sys
import platform
import os
from mattak import __path__ as mattak_path

# this is where pip puts the compiled files
mattak_path = os.path.join(mattak_path[0], 'build/lib/')

loaded = False
loaded_path = None
libmattakName = None
currentPlatform = platform.platform()

def silent_load(what): 
  current = ROOT.gErrorIgnoreLevel
  ROOT.gErrorIgnoreLevel = ROOT.kFatal
  ret = ROOT.gSystem.Load(what)
  ROOT.gErrorIgnoreLevel = current
  return ret
 
if 'macOS' in currentPlatform:
    # print('macOS detected...')
    libmattakName = 'libmattak.dylib'
else: 
    libmattakName = 'libmattak.so'

if not loaded: 
    if not silent_load(libmattakName):
        loaded_path = "LD_LIBRARY_PATH"
        print('Successfully found ' + libmattakName + ' in LD_LIBRARY_PATH')
        loaded = True
    elif not silent_load('build/'+libmattakName):
        print('Successsfully found ' + libmattakName + ' in build')
        loaded_path = "build"
        loaded = True
    elif not silent_load(os.path.join(mattak_path, libmattakName)):
        print('Successsfully found ' + libmattakName + ' in ' + mattak_path)
        loaded_path = mattak_path
        loaded = True
    else:
        for path in sys.path:
            if not silent_load(path + '/mattak/backends/pyroot/'+libmattakName):
                print('Successsfully found ' + libmattakName + ' in ',path)
                loaded_path = path
                loaded = True
                break 

if not loaded: 
    raise Exception('Could not load '+ libmattakName) 


import numpy as np
import uproot
import ROOT
import mattak.Dataset
import time
import os
from os import path

station=11
run=1101
mattakreader = mattak.Dataset.Dataset(
    station, run, 
    # data_dir ='/mnt/pnfs/inbox/',
    data_dir=None,
    backend='pyroot', 
    read_daq_status = False,
    read_run_info = False
)
print ("nentries: " +str( mattakreader.N()))
t0 = time.time()
mattakreader.setEntries((0, None))
mattakreader.eventInfo()
print(f'mattak.Dataset (pyroot backend) took {time.time()-t0:.3f} s')

t0 = time.time()
d =ROOT.mattak.Dataset(station,run) 
for i in range(d.N()): 
    d.setEntry(i)
    d.header() 
print(f'mattak::Dataset (pyroot) took {time.time()-t0:.3f} s')

t0 = time.time()
ROOT.gInterpreter.ProcessLine("mattak::Dataset d (%d,%d); for (int i = 0; i < d.N(); i++) { d.header(); }" % (station,run))
print(f'mattak::Dataset (cling) took {time.time()-t0:.3f} s')

mattakreader = mattak.Dataset.Dataset(
    station, run, 
    # data_dir ='/mnt/pnfs/inbox/',
    data_dir=None,
    backend='uproot', 
    read_daq_status = False, 
    read_run_info = False
)

t0=time.time()
mattakreader.setEntries((0, None))
mattakreader.eventInfo()
print(f'mattak.Dataset (uproot backend) took {time.time()-t0:.3f} s')

t0=time.time()
fname = "%s/station%d/run%d/combined.root" % (os.environ['RNO_G_DATA'], station,run); 
f = None
tree_name = "combined"
if os.path.exists(fname):
    f = uproot.open(fname)
else:
    fname = "%s/station%d/run%d/headers.root" % (os.environ['RNO_G_DATA'], station,run); 
    tree_name = "header" 
    f = uproot.open(fname)

hdr = f[tree_name]['header']
_ = hdr['event_number'].array()
_ = hdr['station_number'].array()
_ = hdr['run_number'].array()
_ = hdr['trigger_time'].array()
_ = hdr['trigger_info'].array()
_ = hdr['pps_num'].array()
_ = hdr['sysclk'].array()
_ = hdr['sysclk_last_pps'].array()
_ = hdr['sysclk_last_last_pps'].array()
_ = hdr['trigger_info/trigger_info.radiant_info.start_windows[24][2]'].array()

print(f'uproot took {time.time()-t0:.3f} s')



import mattak.Dataset
import time
import numpy
import argparse

if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description="Test mattak.")
    argparser.add_argument('--station', type=int, default=21)
    argparser.add_argument('--run', type=int, default=300)
    argparser.add_argument('--data_dir', type=str, default=None)
    argparser.add_argument('--backend', nargs='*', help="Which backend(s) to test", default=['pyroot', 'uproot'])
    args = argparser.parse_args()

    for backend in args.backend:
        print(f">----- Testing backend: {backend} -----<")
        d = mattak.Dataset.Dataset(
            args.station, args.run, data_dir=args.data_dir,
            backend=backend, verbose=True)
        print(d.N())
        print(d.eventInfo())
        print(d.wfs())

        d.setEntries((1,2))
        print(d.eventInfo())
        print(d.wfs())

        mean = 0
        start = time.time()
        for ev in d.iterate():
        #    print ( ev[0].eventNumber,numpy.average(ev[1]))
            mean += numpy.average(ev[1])
        end = time.time()
        print (mean/d.N())
        print ("time:", end-start)




import mattak.Dataset
import time

d = mattak.Dataset.Dataset(23, 999, None, verbose=True)
print(d.N())

tigger_filter = lambda event_info: event_info.triggerType == "FORCE"
time_filter = lambda event_info: 1663968535 < event_info.triggerTime < 1663968971

start = time.time()
d.setEntries((0, d.N()))
for ev in d.iterate(selector=tigger_filter):
    ei, wfs = ev
    print(ei.triggerType)

d.setEntries((0, d.N()))
for ev in d.iterate(selector=time_filter):
    ei, wfs = ev
    print(ei.triggerTime)

end = time.time()
print ("time:", end - start)



